/*
 * Copyright 2001, Olivier Chapuis. No guarantees or warantees or anything
 * are provided or implied in any way whatsoever. Use this program at your
 * own risk. Permission to use this program for any purpose is given,
 * as long as the copyright is kept intact. 
 */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* from the glib code. Thanks! */

#include "config.h"

#if HAVE_NET_HINTS && HAVE_ICONV
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <iconv.h>
#include <errno.h>
#ifdef HAVE_CODESET
#include <locale.h>
#include <langinfo.h>
#endif

#include "libs/fvwmlib.h"
#include "fvwm.h"
#include "externs.h"

#if defined(USE_LIBICONV) && !defined (_LIBICONV_H)
#error libiconv in use but included iconv.h not from libiconv
#endif
#if !defined(USE_LIBICONV) && defined (_LIBICONV_H)
#error libiconv not in use but included iconv.h is from libiconv
#endif

#include "net_name.h"

static char *charset = NULL;
static Bool need_to_get_charset = 1;

extern Atom XA_NET_WM_VISIBLE_NAME;
extern Atom XA_NET_WM_ICON_VISIBLE_NAME;
extern Atom XA_UTF8_STRING;

/***********************************************************************
 * get the charset unsing nl_langinfo
 ***********************************************************************
 */
void get_charset(void)
{
  charset = getenv("CHARSET");

  if (charset)
      return;

#if HAVE_CODESET     
  /* Get the name of the current locale.  */
  if ((setlocale(LC_CTYPE, getenv("LC_CTYPE"))) == NULL)
      fprintf(stderr,
	      "[FVWM]: Can't set locale. Check your $LC_CTYPE or $LANG.\n");

  charset = nl_langinfo(CODESET);
#endif
}

/***********************************************************************
 * iconv
 ***********************************************************************
 */
size_t my_iconv(iconv_t cd, const char **inbuf, size_t *inbytes_left,
		char  **outbuf, size_t  *outbytes_left)
{
  return iconv (cd, inbuf, inbytes_left, outbuf, outbytes_left);
}

/***********************************************************************
 * conversion between charsets
 ***********************************************************************
 */
char *convert_charset(const char *in_charset, const char *out_charset, 
		      const unsigned char *in)
{
  iconv_t cd;
  int have_error = 0;
  size_t nconv;
  size_t insize,outbuf_size,outbytes_remaining,len;
  const char *inptr;
  char *dest;
  char *outp;

  if (in == NULL)
    return NULL;

  inptr = in;

  cd = iconv_open (out_charset, in_charset);
  if (cd == (iconv_t) -1)
  {
    /* Something went wrong.  */
    if (errno == EINVAL)
      fprintf(stderr, "[FVWM]: conversion from `%s' to `%s' not available\n",
	      in_charset,out_charset);
    else
      fprintf (stderr, "[FVWM]: conversion from `%s' to `%s' fail (init)\n",
	       in_charset,out_charset);
    
    /* Terminate the output string.  */
    return NULL;
  }

  len = strlen(in);
  outbuf_size = len + 1;
  outbytes_remaining = outbuf_size - 1;
  insize = len;
  outp = dest = safemalloc(outbuf_size);

 again:

  nconv = my_iconv(cd,(const char **)&inptr,&insize,&outp,&outbytes_remaining);
  if (nconv == (size_t) -1)
    {
      switch (errno)
	{
	case EINVAL:
	  /* Incomplete text, do not report an error */
	  break;
	case E2BIG:
	  {
	    size_t used = outp - dest;

	    outbuf_size *= 2;
	    dest = realloc (dest, outbuf_size);
		
	    outp = dest + used;
	    outbytes_remaining = outbuf_size - used - 1; /* -1 for nul */

	    goto again;
	  }
	default:
	  fprintf(stderr, "[FVWM] Error during conversion from %s to %s:\n %s\n",
		  in_charset,out_charset,strerror (errno));
	  have_error = True;
	  break;
	}
    }


  /* Terminate the output string.  */
  *outp = '\0';
  
  if (iconv_close (cd) != 0)
    fprintf (stderr, "[FVWM]: iconv_close fail");

  if (have_error)
    {
      free (dest);
      return NULL;
    }
  else
    return dest;
}

/***********************************************************************
 * conversion from UTF8 to a "normal" string
 ***********************************************************************
 */
char *convert_utf8(const char *in)
{
  char *out = NULL;
  if (need_to_get_charset)
    get_charset();
  need_to_get_charset = False;

  if (charset != NULL)
    out = convert_charset("UTF8",charset,in);

  return out;
}

/* *********************************************************************
 * XAtom Utils
 * *********************************************************************
 */
void *atom_get(Window win, Atom to_get, Atom type, int *size)
{
  unsigned char	*retval;
  Atom	type_ret;
  unsigned long	 bytes_after, num_ret;
  long length;
  int  format_ret;
  void *data;

  retval = NULL;
  length = 0x7fffffff;
  XGetWindowProperty(dpy, win, to_get, 0,
		     length,
		     False, type,
		     &type_ret,
		     &format_ret,
		     &num_ret,
		     &bytes_after,
		     &retval);

  if ((retval) && (num_ret > 0) && (format_ret > 0))
  {
    data = safemalloc(num_ret * (format_ret >> 3));
    if (data)
      memcpy(data, retval, num_ret * (format_ret >> 3));
    XFree(retval);
    *size = num_ret * (format_ret >> 3);
    return data;
  }
  return NULL;
}

/***********************************************************************
 * Handle the visible names
 ***********************************************************************
 */
void NET_HandleVisibleNames(FvwmWindow *fwin, Bool is_icon_name)
{
  Atom a;
  unsigned char *val;
  unsigned int size = 0;

  if (is_icon_name)
  {
    a = XA_NET_WM_ICON_VISIBLE_NAME;
    NET_FreeVisibleNames(fwin, False, True);
  }
  else
  {
    a = XA_NET_WM_VISIBLE_NAME;
    NET_FreeVisibleNames(fwin, True, False);
  }

  val = atom_get(fwin->w, a, XA_UTF8_STRING, &size);

  if (val == NULL)
  {
    return;
  }

  if (is_icon_name)
  {
     fwin->net_wm_icon_name = convert_utf8((const char *) val);
  }
  else
  {
    fwin->net_wm_name = convert_utf8((const char *) val);
  }
  free(val);
}

/***********************************************************************
 * Free the visible names
 ***********************************************************************
 */
void NET_FreeVisibleNames(FvwmWindow *fwin, Bool nukename, Bool nukeicon)
{
  if (nukename && fwin->net_wm_name != NULL)
  {
    free(fwin->net_wm_name);
    fwin->net_wm_name = NULL;
  }
  if (nukeicon && fwin->net_wm_icon_name != NULL)
  {
    free(fwin->net_wm_icon_name);
    fwin->net_wm_icon_name = NULL;
  }
}
#endif /* HAVE_ICONV && HAVE_NET_HINTS */
